#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 Socktroll server written in python
 Copyright (C) 2008-2010  Patrik Lembke <blambi@chebab.com>

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the Li ccense, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import readline
from blamlib import net

class Socktroll_Server( net.Protocol ):
    def on_connect( self, client ):
        print " * Connection from %s:%d" %( client.host, client.port )
        client.send( "msg server %s\n" % "Welcome to socktroll.py" )
        clients[ client.hash ] = ""

    def on_disconnect( self, client ):
        if not clients.has_key( client.hash ):
            # Workaround for zombies (kicked people)
            return
        
        if clients[ client.hash ] != '':
            print " - %s" % clients[ client.hash ]
            self.broadcast_authed( "- %s\n" % clients[ client.hash ],
                                   client ) # send room join
        else:
            print " * Unauthed client disconnected"

        del clients[ client.hash ]

    def handle( self, client, data ):
        data = data.strip()

        if not client.online:
            print "dead client talking!!! <%s:%d> %s" %( client.host,
                                                         client.port, data )
        
        if not clients.has_key( client.hash ):
            return # someone was removed.

        print "[debug] nick = '%s', data = '%s'" %( clients[client.hash],
                                                    data )
        
        command = data.split( ' ' )[0]
        args = data.split( ' ' )[1:]
        
        # If not authed we vill only have a subset
        if clients[client.hash] == "":
            self.handle_unauthed( client, command, args, client.hash )
        else:
            self.handle_authed( client, command, args, client.hash )

    def handle_unauthed( self, client, command, args, client_id ):
        if command == "nick" and len( args ) == 1:
            self.cmd_nick( client, client_id, args[0] )
        else:
            client.send( "illegal command\n" )
        
    def handle_authed( self, client, command, args, client_id ):
        if command == "nick" and len( args ) == 1:
            self.cmd_nick( client, client_id, args[0] )
        elif command == "msg":
            self.cmd_msg( client, client_id, " ".join( args ) )
        elif command == "action":
            self.cmd_action( client, client_id, " ".join( args ) )
        elif command == "names":
            self.cmd_names( client, client_id, None )
        elif command == "quit":
            self.cmd_quit( client, client_id, None )
        else:
            client.send( "illegal command\n" )
    
    def broadcast_authed( self, data, ignore=None ):
        """set ignore to a client to not send data to it"""
        authed = filter( lambda x: x.hash in clients, self.net.clients )

        for client in authed:
            if client.hash != ignore:
                client.send( data )

    def cmd_nick( self, client, client_id, nick ):
        # On nick, change/auth
        rename = clients[client_id] != ""
        
        # Check if not a word object! ^\w+$
        if len( unicode( nick, "utf-8" ) ) > 20:
            client.send( "no bad\n" )
            return
        
        # check if not taken
        if nick in clients.values() or nick == 'server':
            client.send( "no taken\n" )
            return
        
        # if still here its all fine =)
        if not rename:
            client.send( "ok %s\n" % nick )
            self.broadcast_authed( "+ %s\n" % nick, client ) # send room join
            print " + %s" % nick
            
        else:
            client.send( "rename %s %s\n" %( clients[client_id], nick ) )
            print "rename %s %s" %( clients[client_id], nick )
            self.broadcast_authed( "rename %s %s\n" %( clients[client_id],
                                                     nick ),
                                   client ) # send room rename
            
        clients[client_id] = nick

    def cmd_msg( self, client, client_id, text ):
        self.broadcast_authed( "msg %s %s\n" %( clients[client_id], text ) )

    def cmd_names( self, client, client_id, nan=None ):
        client.send( "names %s\n" % ",".join( clients.values() ) )

    def cmd_quit( self, client, client_id, nan=None ):
        self.net.close_client( client )
        
    def cmd_action( self, client, client_id, text ):
        self.broadcast_authed( "action %s %s\n" %( clients[client_id],
                                                 text ) )

# main
if __name__ == '__main__':
    clients = dict() # ['host:ip'] = nick
    
    print "Socktroll (python server) v0.1.2"
    print
    
    server = net.TCP_Worker( Socktroll_Server, '0.0.0.0', 6000, {} )
    server.start()
    print " * Listning on %s:%d" %( server.host, server.port )
    
    while server.work:
        inp = raw_input( "socktroll> " )

        if inp == 'who' or inp == 'online':
            authed = filter( lambda x: x[1] != '', clients.items() )
            for key, val in authed:
                print "%s\t%s" %( key, val )
            print "%d authed of %d total connections." %( len( authed ),
                                                          len( clients ) )
        elif inp == 'help':
            print "who, online - shows who are online"
            print "help        - this text"
            print "say text..  - says some text as server (this will be to ALL)"
            print "kick nick   - forces nick to leave"
            print "shutdown    - shuts down the server"
            
        elif inp.startswith( 'say' ):
            cmd, message = inp.split( ' ', 1 )
            server.sendtoall( "msg server %s\n" % message )
        elif inp.startswith( 'kick' ):
            cmd, who = inp.split( ' ', 1 )
            if who.find( ' ' ) != -1:
                print "illegal nick..."
            else:
                # find if any, the client that have this nick
                for h_hash, nick in filter( lambda x: x[1] == who,
                                            clients.items() ):

                    prob = filter( lambda c: c.hash == h_hash, server.clients )
                    if len( prob ) != 1:
                        print "More or less then one hit.. this is strange..."
                    else:
                        print "Kicking %s (%s:%d)" %( nick, prob[0].host,
                                                      prob[0].port )
                        del clients[h_hash]
                        server.close_client( prob[0] )
                    
        elif inp == 'shutdown':
            print "Shutting down..."
            server.stop()
            print "Done, bye bye"
        else:
            print "Unrecognized command.."
